#include <sys/wait.h>
#include "ush.h"
#include "history.h"
#include "ush-icmd.h"
#include "ush-var.h"
#include "mstring.h"
#include "mbool.h"
#include "queue.h"
#include "ush-exec.h"

extern queue parsed_commands;
extern history hi;

// private module function prototypes
static void execute_single_command( cmdPtr cmd );

static void
execute_internal_command( char **arg, const int new_stdin, const int new_stdout, const int new_stderr );

static const char *ush_internal_commands[] = { "exit", "history", "pwd", "cd", "set", "alias", NULL };
#define is_internal_command( cmd ) stracmp( cmd, ush_internal_commands )

static void
execute_external_command( char **arg, const int new_stdin, const int new_stdout, const int new_stderr );

static void
execute_external_command_in_bg( char **arg, const int new_stdin, const int new_stdout, const int new_stderr );


// main module function definition
void execute_parsed_commands()
{
    cmdPtr current_command;

    while ( !isempty( parsed_commands ) )
    {
        current_command = dequeue( parsed_commands );

        #if USH_DEBUG
        {
            puts("ush (debug): current parsed command:");
            printf("arguments: ");
            int i=0; while(current_command->arguments[i]) printf("\"%s\" ", current_command->arguments[i++]); puts("");
            printf("standard input redirected: ");
            printf("%s\n", ( current_command->standard_input > 0 ? "yes" : "no" ) );
            printf("standard output redirected: ");
            printf("%s\n", ( current_command->standard_output > 0 ? "yes" : "no" ) );
            printf("standard error redirected: ");
            printf("%s\n", ( current_command->standard_error > 0 ? "yes" : "no" ) );
            printf("background execution: ");
            printf("%s\n", ( current_command->background ? "yes" : "no" ) );
        }
        #endif

        execute_single_command( current_command );
        free_parsed_command( current_command );
    }

    #if USH_DEBUG
        printf("ush (debug): last return code: %s\n", getenv("r"));
    #endif

    return;
}


// private module function definitions

static void execute_single_command( cmdPtr cmd )
{
    if ( cmd->background )
    {
        if ( is_internal_command( cmd->arguments[0] ) )
        {
            fputs("ush: unable to execute internal commands in background\n", stderr);
            return;
        }
        else
            execute_external_command_in_bg( cmd->arguments,
                                            cmd->standard_input,
                                            cmd->standard_output,
                                            cmd->standard_error );
    }
    else // background == false
    {
        if( is_internal_command( cmd->arguments[0] ) )
            execute_internal_command( cmd->arguments,
                                      cmd->standard_input,
                                      cmd->standard_output,
                                      cmd->standard_error );
        else
            execute_external_command( cmd->arguments,
                                      cmd->standard_input,
                                      cmd->standard_output,
                                      cmd->standard_error );
    }
    return;
}

static void
execute_internal_command( char **arg, const int new_stdin, const int new_stdout, const int new_stderr ) // private function definition
{
    #if USH_DEBUG
    {
        printf("ush (debug): executing internal command: \"");
        int i=0;
        while(arg[i]) printf("%s ", arg[i++]);
        puts("\"");
    }
    #endif

    int old_stdin = -1, old_stdout = -1, old_stderr = -1;
    modify_std_streams( new_stdin, new_stdout, new_stderr, &old_stdin, &old_stdout, &old_stderr );

    if( strcmp(arg[0], "exit") == 0 )
    {
        // TO FIX: it doesnt exit with negative numbers
        if( arg[1] && strtol(arg[1], NULL, 0) )
                exit( (int) strtol(arg[1], NULL, 0) );
        else
            exit(0); // clean exit;
    }

    if( strcmp(arg[0], "history") == 0 )
        print_history(hi);

    else if( strcmp(arg[0], "pwd") == 0 )
        puts(read_variable("PWD"));

    else if( strcmp(arg[0], "cd") == 0 )
        cd(arg[1]);

    /* unnecessary - using external echo?
    else if( strcmp(arg[0], "echo") == 0 )
    {
        int i=1;
        while(arg[i]) printf("%s ", arg[i++]);
        puts("");
    }*/

    else if( strcmp(arg[0], "set") == 0 )
        set(arg);

    else if( strcmp(arg[0], "alias") == 0 )
        alias(arg);

    restore_std_streams( old_stdin, old_stdout, old_stderr );

    return;
}

static void
execute_external_command( char **arg, const int new_stdin, const int new_stdout, const int new_stderr ) // private function definition
{
    #if USH_DEBUG
    {
        printf("ush (debug): executing external command: \"");
        int i=0;
        while(arg[i]) printf("%s ", arg[i++]);
        puts("\"");
    }
    #endif

    pid_t c = fork();

    switch ( c )
    {
        case(-1):
        {
            /* fork error */
            fputs("ush: unable to create child process, probably due to insufficient user memory", stderr);
            return;
        }

        case(0):
        {
            /* child-only code */
            modify_std_streams( new_stdin, new_stdout, new_stderr, NULL, NULL, NULL );
            execvp( arg[0], arg );

            /* exec error */
            fprintf( stderr, "ush: %s: command not found\n", arg[0] );
            //fprintf( stderr, "ush: %s: %s\n", arg[0], strerror(errno) );
            exit( USH_EXEC_ERROR );
        }
    }

    int rv;
    waitpid( c, &rv, NULL ); // hangs up and wait until son process finishes
    rv /= 256; // necessary?
    char *srv = itoa(rv); // converts rv integer to string
    write_variable("r", srv); // save termination status
    free(srv);

    return;
}

static void
execute_external_command_in_bg( char **arg, const int new_stdin, const int new_stdout, const int new_stderr ) // private function definition
{
    #if USH_DEBUG
    {
        printf("ush (debug): executing external command in background: \"");
        int i=0;
        while(arg[i]) printf("|%s|", arg[i++]);
        puts("\"");
    }
    #endif

    pid_t c = fork();

    switch ( c )
    {
        case(-1):
        {
            /* fork error */
            fputs("ush: unable to create child process, probably due to insufficient user memory", stderr);
            return;
        }

        case(0):
        {
            /* child-only code */
            modify_std_streams( new_stdin, new_stdout, new_stderr, NULL, NULL, NULL );
            execvp( arg[0], arg );

            /* exec error */
            fprintf( stderr, "ush: %s: command not found\n", arg[0] );
            exit( USH_EXEC_ERROR );
        }
    }

    // prints out PID of the backround command
    printf( "%d\n", c );

    // termination status of the process is catched and written to variable "$r"
    // by the signal handler free_son() registered in ush-init

    return;
}
