#include <fcntl.h>
#include <signal.h>
#include "ush.h"
#include "history.h"
#include "ush-var.h"
#include "queue.h"
#include "ush-init.h"
#include "ush-args.h"
#include "diz.h"

extern char *command;
extern history hi;
extern queue parsed_commands;
extern hash aliases;

// private module function prototypes
static void register_signal_handlers( void );
static void free_son( int signum );
static void initialize_internal_variables( void );
static void read_user_settings ( void );
static void allocate_command_buffer( void );
static void create_history_istance( void );
static void initialize_parsed_commands_queue( void );

// main module funciton definition
void initialize_ush( void )
{
    register_signal_handlers();
    initialize_internal_variables();
    read_user_settings();
    allocate_command_buffer();
    create_history_istance();
    initialize_parsed_commands_queue();
    return;
}


// private module function definitions

static void register_signal_handlers( void )
{
    // ignore SIGINT and SIGQUIT
    signal( SIGINT, SIG_IGN );
    signal( SIGQUIT, SIG_IGN );

    // frees sons once they die
    signal( SIGCHLD, &free_son );

    return;
}

static void free_son( int signum )
{
    int rv, bg_pid;
    while ( (bg_pid = wait( &rv )) > 0 )
    {
        //printf("ush: %d Done\n", bg_pid); // discarded: prints out "Done" to announce termination of a child process
        #if USH_DEBUG
            printf("ush (debug): background process with PID %d terminated\n", bg_pid);
        #endif
    }
    return;
}

static void initialize_internal_variables( void )
{
    // $? = value of the last command run (equivalent of $? in BASH)
    write_variable("r", "0");

    // $$ = current ush's PID
    char *pid = itoa( getpid() );
    write_variable("$", pid);
    free( pid );

    // $p = PATH variable
    write_variable("p", getenv("PATH"));

    // $t = terminal name (e.g. "/dev/tty1")
    // $s = shell starting directory / home directory?
    // $z = shell name with full path (equivalent of $SHELL in BASH)

    return;
}

#define isfile( filename ) ( ! access( filename, F_OK ) )

static void read_user_settings ( void )
{
    char *configuration_file = calloc( sizeof(char), 300 );
    sprintf( configuration_file, "%s/%s", read_variable("HOME"), ".ush_profile" );

    if ( isfile( configuration_file ) ) // if configuration file exists in user's home dir...
    {
        // ...read from it user's configuration file
        FILE *cf = fopen( configuration_file, "r" );
        char *lineptr = calloc(100, sizeof(char)); // if isn't enought is re-allocated by getline
        size_t n = 0;
        char *variable_name = NULL;
        char *variable_value = NULL;
        char **set_declaration[3] = { 0 };

        while ( getline( &lineptr, &n, cf ) != -1 ) // reads configuration file's lines 1 by 1
        {
            if( !(*lineptr == '#') ) // if line doesnt start with an '#' (comment)
            {
                variable_name = lineptr;
                variable_value = strchr( lineptr, '=' );
                *variable_value = '\0'; // split the line in two sub-strings
                variable_value += sizeof(char);
                *(variable_value+strlen(variable_value)-2) = '\0'; // cut off trailing newline

                if( variable_value &&
                    ( strcmp(variable_name, "MAX_CMD_LENGHT") == 0 ) ||
                    ( strcmp(variable_name, "HISTORY_SIZE") == 0 ) ||
                    ( strcmp(variable_name, "MAX_VARIABLE_NAME_LENGHT") == 0 ) )
                {
                    *(variable_value-sizeof(char)) = '='; // undo split strings
                    putenv( strdup( lineptr ) );
                }
                else
                {
                    set_declaration[1] = variable_name;
                    set_declaration[2] = variable_value;
                    set( set_declaration );
                }
            }
        }

        fclose( cf );
        free( lineptr );
    }

    // checks that all system variables are defined correctly
    if( atoi(read_variable("MAX_CMD_LENGHT")) < 10 )
    {
        fputs("ush: too small or undefined MAX_CMD_LENGHT value, restoring default\n", stderr);
        write_variable("MAX_CMD_LENGHT", MAX_CMD_LENGHT_DEFAULT);
    }
    if( atoi(read_variable("HISTORY_SIZE")) < 1 )
    {
        fputs("ush: too small or undefined HISTORY_SIZE value, restoring default\n", stderr);
        write_variable("HISTORY_SIZE", HISTORY_SIZE_DEFAULT);
    }
    if( atoi(read_variable("MAX_VARIABLE_NAME_LENGHT")) < 10 )
    {
        fputs("ush: too small or undefined MAX_VARIABLE_NAME_LENGHT value, restoring default\n", stderr);
        write_variable("MAX_VARIABLE_NAME_LENGHT", MAX_VARIABLE_NAME_LENGHT_DEFAULT);
    }

    // initialize aliases
    aliases = nuovoDizionario();

    write_alias("run", "ush");
    write_alias("abba", "ush");

    free( configuration_file );
}

static void allocate_command_buffer( void )
{
    command = calloc( sizeof(char), atoi(read_variable("MAX_CMD_LENGHT")) );
    if( !command )
    {
        fputs("ush: fatal error: unable to allocate command buffer\nMay be you have specified too big MAX_CMD_LENGHT in your ush.conf?", stderr);
        exit(-1); //exit(ERR_XALLOC);
    }
}

static void create_history_istance( void )
{
    hi = create_history( atoi(read_variable("HISTORY_SIZE")) );
    if( !hi )
    {
        fputs("ush: fatal error: unable to allocate history buffer\nMay be you have specified too big HISTORY_SIZE in your ush.conf?", stderr);
        exit(-1); //exit(ERR_XALLOC);
    }
}

static void initialize_parsed_commands_queue( void )
{
    parsed_commands = createqueue();
}
