#include <unistd.h>
#if _POSIX_C_SOURCE >= 2
     #define USH_USE_WORDEXP
     #include <wordexp.h>
#endif
#include <sys/stat.h>
#include <ctype.h>
#include <fcntl.h>
#include "ush.h"
#include "history.h"
#include "queue.h"
#include "ush-exec.h"
#include "ush-parse.h"
#include "mstring.h"

extern char *command;
extern queue parsed_commands;
extern history hi;

// private module function prototypes
static int check_and_expand_history_cmd( char *cmd );
//static int check_syntax( const char *cmd );
static void check_and_expand_aliases( char *cmd );
static void check_multiple_cmd( const char *cmd, int *new_stdout, bool *bg );
static int check_redirections( const char *cmd, char **arg, int *new_stdin, int *new_stdout, int *new_stderr );
static void do_word_expansion( char *s, char **t );

// main module function definition
void parse_user_input( char *cmd, int in, bool bg )
{
    // skips initial whitespaces
    while( isspace(*cmd) ) cmd++;

    // if empty command returns immediately
    if( !*cmd || *cmd=='\n' ) return;

    // if command starts with "#" it is ignored
    // (used for comments in script file)
    if( *cmd == '#' ) return;

    // create a parsed command istance
    cmdPtr t = create_and_init_parsed_command();
    t->arguments = calloc( sizeof(char*), atoi(read_variable("MAX_CMD_LENGHT"))/2 );
    t->standard_input = in;
    t->background = bg;

    if (
    check_and_expand_history_cmd( cmd )
        == -1 )
    {
        fprintf(stderr, "ush: %s: invalid history position\n", cmd);
        return;
    }

    check_and_expand_aliases( cmd );

    /*if ( check_syntax( cmd ) == -1 )
    {
        fprintf(stderr, "ush: syntax error\n"); //TODO: specificare l'operando mancante?
        return;
    }*/

    check_multiple_cmd( cmd, &(t->standard_output), &(t->background) );

    if (
    check_redirections( cmd, t->arguments, &(t->standard_input),
                                           &(t->standard_output),
                                           &(t->standard_error) )
        == -1 )
    {
        // abort command execution on redirection error
        fprintf(stderr, "ush: %s: %s\n", t->arguments[0], strerror(errno));
        write_variable("r", "1");
        return;
    }

    do_word_expansion( cmd, t->arguments );

    enqueue( parsed_commands, t );

    return;
}


// private module function definitions

static int check_and_expand_history_cmd( char *cmd )
{
    // if the first char of the command is an exclamation mark
    // followed by a valid positive number
    // then reads the command from history
    if ( (*cmd == '!') && (strlen(cmd)>1) && (atoi(cmd+1)>0) )
    {
        int n = atoi( cmd+1 );
        char *newcmd = read_history_elem(n-1, hi);
        if( newcmd )
        {
            strcpy ( cmd, newcmd );
            // TO FIX: it replaces the whole command
            //strrep( cmd, XXX, newcmd );
            //historize(cmd, hi); //re-historize command?
            //fare funzione in history.c: replace_history_elem?
            puts( cmd );
            return 0;
        }
        else
            return -1;
    }
    else
        return 0;
}

static void check_and_expand_aliases( char *cmd )
{
    //TODO: rifare con sscanf?
    // truncate the command string after the first whitespace
    ssize_t i = 0;
    while( *(cmd+i) && !isspace(*(cmd+i)) ) i++;
    *(cmd+i) = '\0';

    // sees if there's an alias for the command
    char *alias_name = strdup( cmd );
    char *alias_value = read_alias( alias_name );

    *(cmd+i) = ' '; // restores the whitespace in command buffer

    if( alias_value )
        strrep( cmd, alias_name, alias_value); // replaces the alias

    free( alias_name );
    return;
}

static int check_syntax( const char *cmd )
{
    // TODO: check that there are not missing/uncoupled ', ", (, {
    return 0;
}

static void check_multiple_cmd( const char *cmd, int *new_stdout, bool *bg )
{
    char *find_last_separator_symbol( char *s ) // private sub-function definition
    {
        char *r = NULL;
        char *t = NULL;
        t = strrstr(s, "|");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, "&");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, ";");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, ")");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, "}");
        if( !isquoted(t, s) && t > r ) r = t;
        /*t = strrstr(s, "||");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, "&&");
        if( !isquoted(t, s) && t > r ) r = t;*/
        return r;
    }

    char *c;
    while ( c = find_last_separator_symbol( cmd ) )
    {
        switch ( *c ) // identify the separator symbol
        {
            case ';':
            {
                parse_user_input( c + sizeof(char), -1, 0 ); // recursively calls the main module function
                break;
            }
            case ')':
            case '}':
            {
                puts("ush: not yet implemented: sub-shell execution");
                break;
            }
            case '|':
            {
                int p[2];
                pipe ( &p ); // creates a pipe
                *new_stdout = p[1]; // current command will use p[1] as stdout
                parse_user_input( c + sizeof(char), p[0], 1 );  // recursively calls the main module function using p[0] as stdin
                break;
            }
            case '&':
            {
                *bg = true;
                parse_user_input( c + sizeof(char), -1, 0 );
                break;
            }
        }
        *c = '\0'; // cut off the command after the separator symbol
    }

    return;
}

static int check_redirections( const char *cmd, char **arg, int *new_stdin, int *new_stdout, int *new_stderr )
{
    char *find_last_redirection_symbol( char *s ) // private sub-function definition
    {
        char *r = NULL;
        char *t = NULL;
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, " < ");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, " > ");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, " 2> ");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, " >> ");
        if( !isquoted(t, s) && t > r ) r = t;
        t = strrstr(s, " 2>> ");
        if( !isquoted(t, s) && t > r ) r = t;
        return r;
    }

    char *c;
    while ( (c = find_last_redirection_symbol( cmd )) )
    {
        *c = '\0'; // cut off the command after the redirection symbol
        c += sizeof(char); // skip empty space before symbol

        // identify redirection symbol and consequently modify standard streams
        if( *new_stdin<0 && strncmp( c, "< ", 2 )==0 )
        {
            do_word_expansion( (c+sizeof(char)), arg );
            *new_stdin = open ( arg[0], O_RDONLY );
            if ( *new_stdin < 0 )
                return -1;
            else
                continue;
        }
        else if( *new_stdout<0 && strncmp( c, "> ", 2 )==0 )
        {
            do_word_expansion( (c+sizeof(char)), arg );
            *new_stdout = creat( arg[0], S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
            if ( *new_stdout < 0 )
                return -1;
            else
                continue;
        }
        else if( *new_stderr<0 && strncmp( c, "2> ", 3 )==0 )
        {
            do_word_expansion( (c+3*sizeof(char)), arg );
            *new_stderr = creat( arg[0], S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
            if ( *new_stderr < 0 )
                return -1;
            else
                continue;
        }
        else if( *new_stdout<0 && strncmp( c, ">> ", 3 )==0 )
        {
            do_word_expansion( (c+3*sizeof(char)), arg );
            *new_stdout = open ( arg[0], O_WRONLY | O_APPEND | O_CREAT );
            if ( *new_stdout < 0 )
                return -1;
            else
                continue;
        }
        else if( *new_stderr<0 && strncmp( c, "2>> ", 4 )==0 )
        {
            do_word_expansion( (c+4*sizeof(char)), arg );
            *new_stderr = open ( arg[0], O_WRONLY | O_APPEND | O_CREAT );
            if ( *new_stderr < 0 )
                return -1;
            else
                continue;
        }
    }

    return 0;
}

static void do_word_expansion( char *s, char **t )
#ifdef USH_USE_WORDEXP
// implementation using wordexp (avaiable only in POSIX.2)
{
    wordexp_t r;

    switch( wordexp (s, &r, 0) )
    {
        case 0:
            break;
        case WRDE_NOSPACE:
            /* If the error was WRDE_NOSPACE,
            then perhaps part of the result was allocated. */
            wordfree (&r);
        default:
            return; // TODO: return errors?
    }

    int i = 0;
    while ( r.we_wordv[i] )
    {
        t[i] = strdup( r.we_wordv[i] );
        i++;
    }

    wordfree(&r);
}
#else
// implementation using POSIX.1-only functions (NOT YET FULLY FUNCTIONAL)
{
    char *read_next_variable_name( const char *t ) // private sub-function definition
    {

        char *b = strchr(t, '$');
        if (!b) return NULL;

        // TODO: check for quoting
        //if( isquoted(b, t) )
            b = strchr(++b, '$');

        int max_variable_name_lenght =
            atoi( read_variable("MAX_VARIABLE_NAME_LENGHT") );

        char *n = calloc( sizeof(char), max_variable_name_lenght );

        int l = 0;
        while( *b != '\0' &&
               *b != ' ' &&
               l<max_variable_name_lenght )
        {
            *(n+l) = *b;
            b++;
            l++;
        }
        return n;
    }

    // tokenizes the input command using whitespaces as separators
    t[0] = strdup( strtok( s, " \t" ) );
    int i = 1;
    char *k;
    while ( k = strtok( NULL, " \t" )  )
    {
        t[i] = strdup(k);
        i++;
    }
    return;

    char *name = NULL;
    char *value = NULL;

    while ( name = read_next_variable_name( s ) )
    {
        value = read_variable( name+1 );

        strrep( s, name, value );

        free( name );
        free( value );
    }
}
#endif
